# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__version__ = '0.14.8'


from . import algebra  # noqa
from . import architectures  # noqa
from . import callbacks  # noqa
from . import classification_metrics  # noqa
from . import distances  # noqa
from . import evaluators  # noqa
from . import layers  # noqa
from . import losses  # noqa
from . import indexer  # noqa
from . import matchers  # noqa
from . import models  # noqa
from . import retrieval_metrics  # noqa
from . import samplers  # noqa
from . import search  # noqa
from . import stores  # noqa
from . import training_metrics  # noqa
from . import types  # noqa
from . import utils  # noqa
from . import visualization  # noqa
