# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matchers define the classification matching strategy when using
similarity models to match query examples to the classes of indexed
examples."""
# Match Types
from .classification_match import ClassificationMatch  # noqa
from .match_majority_vote import MatchMajorityVote  # noqa
from .match_nearest import MatchNearest  # noqa
from .utils import make_classification_matcher  # noqa
